package hu.bme.mit.book.gui;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Iterator;

import hu.bme.mit.books.manager.IBookManager;
import hu.bme.mit.books.model.Book;

import org.eclipse.core.databinding.Binding;
import org.eclipse.core.databinding.DataBindingContext;
import org.eclipse.core.databinding.UpdateValueStrategy;
import org.eclipse.core.databinding.beans.BeanProperties;
import org.eclipse.core.databinding.beans.BeansObservables;
import org.eclipse.core.databinding.conversion.IConverter;
import org.eclipse.core.databinding.observable.list.WritableList;
import org.eclipse.core.databinding.observable.value.IObservableValue;
import org.eclipse.core.databinding.validation.IValidator;
import org.eclipse.core.databinding.validation.ValidationStatus;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.jface.databinding.swt.SWTObservables;
import org.eclipse.jface.databinding.viewers.ViewerSupport;
import org.eclipse.jface.databinding.viewers.ViewersObservables;
import org.eclipse.jface.dialogs.IMessageProvider;
import org.eclipse.jface.fieldassist.ControlDecoration;
import org.eclipse.jface.fieldassist.FieldDecoration;
import org.eclipse.jface.fieldassist.FieldDecorationRegistry;
import org.eclipse.jface.viewers.ColumnWeightData;
import org.eclipse.jface.viewers.IStructuredSelection;
import org.eclipse.jface.viewers.TableLayout;
import org.eclipse.jface.viewers.TableViewer;
import org.eclipse.jface.viewers.TableViewerColumn;
import org.eclipse.swt.SWT;
import org.eclipse.swt.graphics.Color;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.Button;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.Table;
import org.eclipse.swt.widgets.Text;
import org.eclipse.ui.PlatformUI;
import org.eclipse.ui.forms.widgets.ColumnLayout;
import org.eclipse.ui.forms.widgets.Form;
import org.eclipse.ui.forms.widgets.FormToolkit;
import org.eclipse.ui.forms.widgets.Section;
import org.eclipse.ui.part.ViewPart;

public class BookDetailsView extends ViewPart implements IServiceInterface {

	TableViewer viewer;

	private final class StringToIntConverter implements IConverter {
		@Override
		public Object getToType() {
			return Integer.class;
		}

		@Override
		public Object getFromType() {
			return String.class;
		}

		@Override
		public Object convert(Object fromObject) {
			if (fromObject == null)
				return "";
			return Integer.parseInt((String) fromObject);
		}
	}

	private final class IntToStringConverter implements IConverter {
		@Override
		public Object getToType() {
			return String.class;
		}

		@Override
		public Object getFromType() {
			return Integer.class;
		}

		@Override
		public Object convert(Object fromObject) {
			if (fromObject == null)
				return null;
			return ((Integer) fromObject).toString();
		}
	}

	private final class PagenumValidator implements IValidator {

		ControlDecoration decoration;

		public PagenumValidator(ControlDecoration controlDecoration) {
			this.decoration = controlDecoration;
		}

		@Override
		public IStatus validate(Object value) {
			int num = 0;
			boolean valid = true;
			try {
				num = Integer.parseInt((String) value);
			} catch (NumberFormatException e) {
				valid = false;
			}
			valid = valid && (num > 0);
			if (valid) {
				decoration.hide();
				return ValidationStatus.ok();
			} else {
				decoration.show();
				decoration
						.setDescriptionText("The page number has to be a positive number.");
				return ValidationStatus
						.error("The page number has to be a positive number.");
			}
		}

	}

	private final class StatusToColorConverter implements IConverter {

		@Override
		public Object convert(Object fromObject) {
			IStatus status = (IStatus) fromObject;
			if (status.getSeverity() == IStatus.ERROR)
				return Display.getCurrent().getSystemColor(SWT.COLOR_RED);
			return Display.getCurrent().getSystemColor(SWT.COLOR_WHITE);
		}

		@Override
		public Object getFromType() {
			return IStatus.class;
		}

		@Override
		public Object getToType() {
			return Color.class;
		}

	}

	FormToolkit toolkit;
	Form form;

	private Text author;
	private Text title;
	private Text pagenum;
	private Button read;

	private IBookManager manager;
	private WritableList input;

	@Override
	public void createPartControl(Composite parent) {
		toolkit = new FormToolkit(parent.getDisplay());
		form = toolkit.createForm(parent);
		toolkit.decorateFormHeading(form);
		form.setText("Book Demo");
		Composite body = form.getBody();
		body.setLayout(new ColumnLayout());

		// Booklist section
		Section booksSection = toolkit.createSection(body, Section.DESCRIPTION
				| Section.TITLE_BAR | Section.EXPANDED);
		booksSection.setText("Books");
		Table table = toolkit.createTable(booksSection, SWT.FULL_SELECTION);
		booksSection.setClient(table);
		viewer = new TableViewer(table);
		TableViewerColumn column = new TableViewerColumn(viewer, SWT.LEFT);
		TableLayout tableLayout = new TableLayout();
		tableLayout.addColumnData(new ColumnWeightData(100, true));
		table.setLayout(tableLayout);
		table.setHeaderVisible(false);
		table.setLinesVisible(true);

		// Book details section
		Section detailsSection = toolkit.createSection(body,
				Section.DESCRIPTION | Section.TITLE_BAR | Section.EXPANDED);
		detailsSection.setText("Book Details");
		detailsSection.setDescription("Book Details");
		Composite detailsComposite = toolkit.createComposite(detailsSection);
		detailsSection.setClient(detailsComposite);
		detailsComposite.setLayout(new GridLayout(2, true));
		toolkit.createLabel(detailsComposite, "Author");
		author = toolkit.createText(detailsComposite, "Author name");
		toolkit.createLabel(detailsComposite, "Title");
		title = toolkit.createText(detailsComposite, "Book title");
		toolkit.createLabel(detailsComposite, "PageNum");
		pagenum = toolkit.createText(detailsComposite, "Number of pages");
		read = toolkit.createButton(detailsComposite, "Read", SWT.CHECK);

		// Layouting
		GridData gd = new GridData(GridData.FILL_HORIZONTAL);
		gd.widthHint = 150;
		author.setLayoutData(gd);
		title.setLayoutData(gd);
		pagenum.setLayoutData(gd);

		// Loading content
		ServiceManager.registerServiceListener(this);
		loadManager();

		// Control Decoration
		ControlDecoration controlDecoration = new ControlDecoration(pagenum,
				SWT.LEFT | SWT.TOP);
		controlDecoration.setDescriptionText("hoverText");
		FieldDecoration fieldDecoration = FieldDecorationRegistry.getDefault()
				.getFieldDecoration(FieldDecorationRegistry.DEC_ERROR);
		controlDecoration.setImage(fieldDecoration.getImage());

		// Data binding
		IObservableValue selection = ViewersObservables
				.observeSingleSelection(viewer);
		IObservableValue authorObservable = BeansObservables
				.observeDetailValue(selection, "author", String.class);
		IObservableValue titleObservable = BeansObservables.observeDetailValue(
				selection, "title", String.class);
		IObservableValue pageNumObservable = BeansObservables
				.observeDetailValue(selection, "pagenum", Integer.class);
		IObservableValue readObservable = BeansObservables.observeDetailValue(
				selection, "read", Boolean.class);
		DataBindingContext dbc = new DataBindingContext();
		UpdateValueStrategy updateStrategy = new UpdateValueStrategy(false,
				UpdateValueStrategy.POLICY_UPDATE);
		dbc.bindValue(SWTObservables.observeText(title, SWT.Modify),
				titleObservable, updateStrategy, updateStrategy);
		dbc.bindValue(SWTObservables.observeText(author, SWT.Modify),
				authorObservable);
		Binding pagenumBind = dbc.bindValue(SWTObservables.observeText(pagenum,
				SWT.Modify), pageNumObservable, new UpdateValueStrategy()
				.setConverter(new StringToIntConverter()).setAfterGetValidator(
						new PagenumValidator(controlDecoration)),
				new UpdateValueStrategy()
						.setConverter(new IntToStringConverter()));
		dbc.bindValue(SWTObservables.observeSelection(read), readObservable);

		dbc.bindValue(SWTObservables.observeBackground(pagenum), pagenumBind
				.getValidationStatus(), null, new UpdateValueStrategy()
				.setConverter(new StatusToColorConverter()));

	}

	public void loadManager() {
		PlatformUI.getWorkbench().getDisplay().asyncExec(new Runnable() {

			@Override
			public void run() {
				manager = ServiceManager.getManager();
				if (manager == null) {
					form.setMessage("Service not available",
							IMessageProvider.ERROR);
					input = new WritableList(new ArrayList<Book>(),
							Book.class);
					viewer.setInput(input);
				} else {
					if (viewer.getInput() != null) {
						viewer.setInput(new WritableList(manager.getAllBooks(),
								Book.class));
					} else {
						ViewerSupport.bind(viewer, new WritableList(manager
								.getAllBooks(), Book.class), BeanProperties
								.value(Book.class, "description"));
					}
					form.setMessage("");
				}
			}
		});
	}

	@Override
	public void setFocus() {
		form.setFocus();

	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.eclipse.ui.part.WorkbenchPart#dispose()
	 */
	@Override
	public void dispose() {
		toolkit.dispose();
		ServiceManager.unregisterServiceListener(this);
	}

	@Override
	public void serviceStarted() {
		loadManager();
	}

	@Override
	public void serviceStopped() {
		loadManager();
	}

	public Book addBook(String author, String title, int pageNum, boolean isRead) {
		//XXX book list changes are not propagated back into the service
		Book book = new Book();
		book.setAuthor(author);
		book.setTitle(title);
		book.setPagenum(pageNum);
		book.setRead(isRead);
		input.add(book);
		return book;
	}
	
	public void removeSelectedBook() {
		//XXX book list changes are not propagated back into the service
		IStructuredSelection selection = (IStructuredSelection) viewer.getSelection();
		Iterator iterator = selection.iterator();
		while (iterator.hasNext()) {
			Book book = (Book) iterator.next();
			input.remove(book);
		}
	}

}
